/*
 * Copyright 2015-2017, Thomas Maier-Komor
 * LICENSE: GPLv3, see file LICENSE for details
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "log.h"
#include "support.h"

#include <vector>


using namespace std;


struct Disk
{
	size_t index;		// index in hostscope::disks
	uint64_t ts;		// timestamp of last update
	uint64_t size;		// kB on disk reported by /sys/block/.../size
	uint64_t nrd		// num read issued
		, nmrd		// num merged reads
		, nsecrd	// num sectors read
		, msrd		// ms spent reading
		, nwr		// num writes
		, nmwr		// num merged writes
		, nsecwr	// num sectors written
		, mswr		// ms spent writing
		, niop		// num I/Os in progress
		, msio		// ms spent doing I/O
		, wmsio;	// weighted num ms spent doing I/O
	bool removable;
	string devname;

	Disk(const char *dname, uint64_t s, size_t i)
	: index(i)
	, ts(0)
	, size(s)
	, nrd(0)
	, nmrd(0)
	, nsecrd(0)
	, msrd(0)
	, nwr(0)
	, nmwr(0)
	, nsecwr(0)
	, mswr(0)
	, niop(0)
	, msio(0)
	, wmsio(0)
	, removable(false)
	, devname(dname)
	{ }
};

static vector<Disk> Disks;


static void get_disk_size(const char *devname, HostScope *t, int idx = -1)
{
	TimeStat(__FUNCTION__);
	char spath[64] = "/sys/block/";
	char *s = stpcpy(spath+11,devname);
	strcpy(s,"/size");
	int sf = open(spath,O_RDONLY);
	if (-1 == sf) {
		warning("unable to open file %s: %s\n",spath,strerror(errno));
		return;
	}
	char buf[256];
	int n = read(sf,buf,sizeof(buf));
	assert(n < (int)sizeof(buf));
	close(sf);
	if (n == -1) {
		warning("unable to read file %s: %s\n",spath,strerror(errno));
		return;
	}
	buf[n] = 0;
	long long ds = strtoll(buf,0,10);
	ds >>= 1;
	HostScope_Disk *td = 0;
	//assert(Disks.size() == t->disks_size());
	if (idx == -1) {
		size_t index = Disks.size();
		Disks.push_back(Disk(devname,ds,index));
		td = t->add_disks();
		td->set_devname(devname);
		strcpy(s,"/removable");
		int rf = open(spath,O_RDONLY);
		if (-1 != rf) {
			n = read(rf,buf,sizeof(buf));
			close(rf);
			assert(n < (int)sizeof(buf));
			if (n > 0) {
				buf[n] = 0;
				Disks.back().removable = strtol(buf,0,10) != 0;
			}
		}
	} else {
		Disks[idx].size = ds;
		assert(Disks[idx].devname == devname);
		for (size_t i = 0; i < (size_t)t->disks_size(); ++i) {
			HostScope_Disk *x = t->mutable_disks(i);
			if (0 == strcmp(x->devname().c_str(), devname)) {
				td = x;
				break;
			}
		}
	}
	assert(td != 0);
	td->set_ktotal(ds);
	dbug("device %s, size %luMB\n",devname,ds>>11);
}


static void init_disks(HostScope *t)
{
	TimeStat(__FUNCTION__);
	static DIR *dir = opendir("/sys/block");
	if (0 == dir) {
		warnonce("unable to open directory /sys/block: %s\n",strerror(errno));
		return;
	}
	for (struct dirent *de = readdir(dir); de; de = readdir(dir)) {
		if (de->d_name[0] == '.') {
			de = readdir(dir);
			continue;
		}
		if (0 == strncmp(de->d_name,"loop",4)) 
			if ((de->d_name[4] >= '0') && (de->d_name[4] <= '9'))
				continue;
		if (0 == strncmp(de->d_name,"ram",3)) 
			if ((de->d_name[3] >= '0') && (de->d_name[3] <= '9'))
				continue;
		dbug("check disk %s\n",de->d_name);
		vector<Disk>::iterator i(Disks.begin()), e(Disks.end());
		for (; i != e; ++i) {
			if (i->devname == de->d_name)

				break;
		}
		if (i == e)
			get_disk_size(de->d_name,t);
	}
	rewinddir(dir);
}


void read_diskinfo(HostScope *t)
{
	static unsigned i = 0;
	TimeStat(__FUNCTION__);
	t->clear_disks();
	if ((i % 5) == 0)
		init_disks(t);
	static int fd = open("/proc/diskstats",O_RDONLY);
	if (fd < 0) {
		warnonce("unable to open /proc/diskstats: %s\n",strerror(errno));
		return;
	}
	char buf[4096];
	int n = pread(fd,buf,sizeof(buf),0);
	assert(n < (int)sizeof(buf));
	if (n < 0) {
		warnonce("unable to read /proc/diskstats: %s\n",strerror(errno));
		return;
	}
	buf[n] = 0;
	uint64_t ts = get_us_time();
	char *at = buf;
	for (;at && *at;) {
		uint64_t nrd, nmrd, nsecrd, msrd, nwr, nmwr, nsecwr, mswr, niop, msio, wmsio;
		char dev[32];
		int f = sscanf(at,"%*u %*u %31s %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 " %" SCNu64 ""
				, dev
				, &nrd
				, &nmrd
				, &nsecrd
				, &msrd
				, &nwr
				, &nmwr
				, &nsecwr
				, &mswr
				, &niop
				, &msio
				, &wmsio);
		if (f == 0)
			return;
		at = strchr(at,'\n');
		if (at)
			++at;
		if ((nrd == 0) && (nwr == 0))
			continue;
		Disk *d = 0;
		for (vector<Disk>::iterator i(Disks.begin()), e(Disks.end()); i != e; ++i) {
			if (0 == strcmp(dev, i->devname.c_str())) {
				d = &(*i);
				break;
			}
		}
		if (d == 0)
			continue;
		if (d->ts != 0) {
			double dt = (double)((ts - d->ts)/1000);
			HostScope_Disk *td = t->add_disks();
			td->set_devname(d->devname);
			td->set_nread(dbl2fxptE4((double)(nrd-d->nrd)/dt*1E3));
			td->set_nwrite(dbl2fxptE4((double)(nwr-d->nwr)/dt*1E3));
			td->set_ktotal(d->size);
			td->set_kread(dbl2fxptE4((double)((nsecrd-d->nsecrd)>>1)/(double)(msrd-d->msrd)));
			td->set_kwrite(dbl2fxptE4((double)((nsecwr-d->nsecwr)>>1)/(double)(mswr-d->mswr)));
			// percent busy can become > 100% because of
			// different clocks. the time of msio sampling
			// is unavailable, therefore we have to estimate
			// it, be reading the monotonic clock...
			double percbusy = (double)(msio-d->msio)/dt;
			if (percbusy > 1)
				percbusy = 1;
			td->set_perc_busy(frac2fxptE4(percbusy));
			td->set_opsqueued(niop);
		}
		if (d->removable)
			get_disk_size(dev,t,d->index);
		d->ts = ts;
		d->nrd = nrd;
		d->nmrd = nmrd;
		d->nsecrd = nsecrd;
		d->msrd = msrd;
		d->nwr = nwr;
		d->nmwr = nmwr;
		d->nsecwr = nsecwr;
		d->mswr = mswr;
		d->niop = niop;
		d->msio = msio;
		d->wmsio = wmsio;
	}
	++i;
}


